package rice.pastry.socket;

import java.io.IOException;
import java.net.*;
import java.util.Iterator;
import java.util.Set;

import net.sbbi.upnp.impls.InternetGatewayDevice;
import net.sbbi.upnp.messages.ActionResponse;
import net.sbbi.upnp.messages.UPNPResponseException;
import rice.Continuation.ExternalContinuation;
import rice.environment.Environment;
import rice.environment.logging.Logger;
import rice.environment.params.Parameters;
import rice.pastry.socket.messaging.IPAddressRequestMessage;
import rice.pastry.socket.messaging.IPAddressResponseMessage;

public class SocketNATHandler extends StubNATHandler implements NATHandler {
  public static final int ALWAYS = 1;
  public static final int PREFIX_MATCH = 2;
  public static final int NEVER = 3;
  public static final int OVERWRITE = 1;
  public static final int USE_DIFFERENT_PORT = 2;
  public static final int FAIL = 3;

  public SocketNATHandler(Environment env, InetSocketAddress localAddress, InetSocketAddress proxyAddress) {
    super(env,localAddress,proxyAddress);
  }
  
  boolean searchedForFireWall = false;
  InternetGatewayDevice fireWall;
  InetAddress fireWallExternalAddress;
  
  public synchronized InetAddress findFireWall(InetAddress bindAddress) throws IOException {
    NetworkInterface ni = NetworkInterface.getByInetAddress(bindAddress);
    if (searchedForFireWall) return fireWallExternalAddress;
    searchedForFireWall = true;
    int discoveryTimeout = environment.getParameters().getInt("nat_discovery_timeout");
    
    InternetGatewayDevice[] IGDs = InternetGatewayDevice.getDevices(discoveryTimeout); 
    // use this code with the next version of sbbi's upnp library, it will only search for the firewall on the given NetworkInterface
//    InternetGatewayDevice[] IGDs = InternetGatewayDevice.getDevices(discoveryTimeout, 
//        Discovery.DEFAULT_TTL,
//        Discovery.DEFAULT_MX,
//        ni);
    if ( IGDs != null ) {
      // no idea how to handle this if there are 2 firewalls... handle the first one
      // if they have that interesting of a network, then they know what they are doing
      // and can configure port forwarding      
      fireWall = IGDs[0];
      try {
        fireWallExternalAddress = InetAddress.getByName(fireWall.getExternalIPAddress());
      } catch (UPNPResponseException e) {
        throw new IOException(e.getMessage());
      } 
    } else {
      throw new IOException("Could not find firewall.  Please enable UPnP on firewall, or set 'find_firewall_policy = never'");
    }
    return fireWallExternalAddress;
  }
  
  
  /**
   * Returns the external port to use based on querying the firewall for both 
   * TCP and UDP
   * 
   * @param internal // the internal port we expect it to be mapped to 
   * @param external // the first port to try, will search by incrementing
   * @return
   * @throws IOException
   * @throws UPNPResponseException
   */
  int findPortTries = 0;
  public int findAvailableFireWallPort(int internal, int external) throws IOException  {
    try {
      findPortTries++;
      if (findPortTries > environment.getParameters().getInt("nat_find_port_max_tries")) 
          throw new IOException("Couldn't find available port on firewall");
      if (logger.level <= Logger.FINE) logger.log("findFireWallPort("+internal+","+external+")");
      String appName = environment.getParameters().getString("nat_app_name");
      ActionResponse response = null;
      response = fireWall.getSpecificPortMappingEntry(null,
          external,
          "TCP");
      if (checkSpecificPortMappingEntryResponse(response, internal, external, "TCP", appName)) {
        // continue
      } else {
        return findAvailableFireWallPort(internal, external+1);
      }
          
      response = fireWall.getSpecificPortMappingEntry(null,
          external,
          "UDP"); 
      if (checkSpecificPortMappingEntryResponse(response, internal, external, "UDP", appName)) {
        // continue
      } else {
        return findAvailableFireWallPort(internal, external+1);
      }
      return external;
    } catch (UPNPResponseException e) {
      throw new IOException(e.getMessage());
    }    
  }
  
  public static final int MAX_PORT = 65535;
  
  /**
   * We return true if the response matches our app exactly, or if the entries 
   * are invalid (because some firewalls do this instad of returning an error 
   * that there is no entry.)
   * 
   * @param response
   * @param internal
   * @param external
   * @param type TCP or UDP
   * @return
   */
  private boolean checkSpecificPortMappingEntryResponse(ActionResponse response, int internal, int external, String type, String app) {
    if (response == null) return true; 

    if (logger.level <= Logger.FINEST) {
      Set s = response.getOutActionArgumentNames();
      Iterator i = s.iterator();
      while(i.hasNext()) {
        String key = (String)i.next();
        String val = response.getOutActionArgumentValue(key);
        System.out.println("  "+key+" -> "+val); 
      }
    }
    
    boolean ret = true;
    //    NewInternalPort->9001
    String internalPortStr = response.getOutActionArgumentValue("NewInternalPort");
    if (internalPortStr != null) {
      try {
        int internalPort = Integer.parseInt(internalPortStr); 
        if (internalPort > MAX_PORT) {
          if (logger.level < Logger.WARNING) logger.log(
              "Warning, NAT "+
              fireWall.getIGDRootDevice().getModelName()+
              " returned an invalid value for entry NewInternalPort.  Expected an integer less than "+
              MAX_PORT+", got: "+
              internalPort+".  Query "+external+":"+type+"... overwriting entry.");
          return true; // a bogus entry from the firewall, 
        }
        if (internalPort != internal) {
          if (logger.level <= Logger.FINER) logger.log("internalPort("+internalPort+") != internal("+internal+")");
          ret = false; // but don't return yet, maybe it will become more clear with another clue
        }
      } catch (NumberFormatException nfe) {
        // firewall bug, assume we can overwrite
        if (logger.level < Logger.WARNING) logger.log(
            "Warning, NAT "+
            fireWall.getIGDRootDevice().getModelName()+
            " returned an invalid value for entry NewInternalPort.  Expected an integer, got: "+
            internalPortStr+".  Query "+external+":"+type+"... overwriting entry.");
        return true; 
      }
    }
    
    //    NewPortMappingDescription->freepastry
    String appName = response.getOutActionArgumentValue("NewPortMappingDescription");
    if (appName != null) {
      // this is in case the app name is just a bunch of spaces
      String tempName = appName.replaceAll(" ","");
      if ((tempName.equals("")) || appName.equalsIgnoreCase(app)) {
        // do nothing yet
      } else {
        if (logger.level <= Logger.FINER) logger.log("appName("+appName+") != app("+app+")");
        ret = false; 
      }
    }

    //    NewInternalClient->192.168.1.64
    String newInternalClientString = response.getOutActionArgumentValue("NewInternalClient");
    if (newInternalClientString == null) {
      if (logger.level < Logger.WARNING) logger.log(
          "Warning, NAT "+
          fireWall.getIGDRootDevice().getModelName()+
          " returned no value for entry NewInternalClient.  Expected an IP address, got: "+
          newInternalClientString+".  Query "+external+":"+type+"... overwriting entry.");
      return true;
    }
    try {
      InetAddress client = InetAddress.getByName(newInternalClientString);            
      if (!client.equals(localAddress)) {
        if (logger.level <= Logger.FINER) logger.log("client("+client+") != localAddress("+localAddress+")");
        ret = false; 
      }
    } catch (Exception e) {
      if (logger.level < Logger.WARNING) logger.log(
          "Warning, NAT "+
          fireWall.getIGDRootDevice().getModelName()+
          " returned an invalid value for entry NewInternalClient.  Expected an IP address, got: "+
          newInternalClientString+".  Query "+external+":"+type+"... overwriting entry.");
      return true;
    }
    
    //    NewEnabled->1
    String enabledString = response.getOutActionArgumentValue("NewEnabled");      
    if (enabledString != null) {
      try {
        int enabled = Integer.parseInt(enabledString); 
        if (enabled == 0) {
          if (logger.level < Logger.FINE) logger.log(
              "Warning, NAT "+
              fireWall.getIGDRootDevice().getModelName()+
              " had an existing rule that was disabled, implicitly overwriting.  "+
              "Query "+external+":"+type+"."+
              "\n  NewInternalPort -> "+internalPortStr+
              "\n  NewPortMappingDescription -> "+appName+
              "\n  NewInternalClient -> "+newInternalClientString+
              "\n  NewEnabled -> "+enabledString);            
          return true; // the current rule is not used, go ahead and overwrite
        } else if (enabled == 1) {
          // nothing, depend on previous settings of ret to determine what to do
        } else {
          if (logger.level < Logger.WARNING) logger.log(
              "Warning, NAT "+
              fireWall.getIGDRootDevice().getModelName()+
              " returned an invalid value for entry NewEnabled.  Expected 0 or 1, got: "+
              enabled+".  Query "+external+":"+type+"... overwriting entry.");
          return true; // a bogus entry from the firewall, 
        } 
      } catch (NumberFormatException nfe) {
        // firewall bug, assume we can overwrite
        if (logger.level < Logger.WARNING) logger.log(
            "Warning, NAT "+
            fireWall.getIGDRootDevice().getModelName()+
            " returned an invalid value for entry NewEnabled.  Expected 0 or 1, got: "+
            enabledString+".  Query "+external+":"+type+"... overwriting entry.");
        return true; 
      }
    } else {
      // router didn't specify enable string, no info, do nothing
    }

    if (ret == false) {
      if (logger.level < Logger.INFO) logger.log(
          "Warning, NAT "+
          fireWall.getIGDRootDevice().getModelName()+
          " had an existing rule, trying different port.  "+
          "Query "+external+":"+type+"."+
          "\n  NewInternalPort -> "+internalPortStr+
          "\n  NewPortMappingDescription -> "+appName+
          "\n  NewInternalClient -> "+newInternalClientString+
          "\n  NewEnabled -> "+enabledString);            
    }        
    return ret;
  }
  
  public void openFireWallPort(int local, int external) throws IOException {
    try {
      boolean mapped = true;
      mapped = fireWall.addPortMapping(
          environment.getParameters().getString("nat_app_name"), 
          null, local, external,
          localAddress.getAddress().getHostAddress(), 0, "TCP" );
      if (!mapped) throw new IOException("Could not set firewall TCP port forwarding from external:"+fireWallExternalAddress+":"+external+" -> local:"+localAddress+":"+local);
      mapped = fireWall.addPortMapping(
          environment.getParameters().getString("nat_app_name"), 
          null, local, external,
          localAddress.getAddress().getHostAddress(), 0, "UDP" );
      if (!mapped) throw new IOException("Could not set firewall UDP port forwarding from external:"+fireWallExternalAddress+":"+external+" -> local:"+localAddress+":"+local);
    } catch (UPNPResponseException e) {
      throw new IOException(e.getMessage());
    } 
  }

  protected int getFireWallPolicyVariable(String key) {
    String val = environment.getParameters().getString(key);
    if (val.equalsIgnoreCase("prefix")) return PREFIX_MATCH;
    if (val.equalsIgnoreCase("change")) return USE_DIFFERENT_PORT;
    if (val.equalsIgnoreCase("never")) return NEVER;
    if (val.equalsIgnoreCase("overwrite")) return OVERWRITE;
    if (val.equalsIgnoreCase("always")) return ALWAYS;
    if (val.equalsIgnoreCase("fail")) return FAIL;
    throw new RuntimeException("Unknown value "+val+" for "+key);
  }
  
  /**
   * @return true if ip address matches firewall prefix
   */
  public boolean localAddressIsProbablyNatted() {
    String ip = localAddress.getAddress().getHostAddress();
    String nattedNetworkPrefixes = environment.getParameters().getString("nat_network_prefixes");
    
    String[] nattedNetworkPrefix = nattedNetworkPrefixes.split(";");
    for (int i = 0; i < nattedNetworkPrefix.length; i++) {
      if (ip.startsWith(nattedNetworkPrefix[i])) {
        return true;
      }
    }
    return false;
  }


  public boolean findFireWallIfNecessary() throws IOException {
    switch(getFireWallPolicyVariable("nat_search_policy")) {
      case NEVER:
        return false;
      case PREFIX_MATCH:
        if (!localAddressIsProbablyNatted()) return false;
      case ALWAYS:
      default:
        findFireWall(localAddress.getAddress());
        return true;
    }
  }

  public boolean usingNAT() {
    return !pAddress.getAddress().equals(localAddress.getAddress());
  }

  public InetSocketAddress externalAddress() throws IOException {
    int port = localAddress.getPort();
    
    if (findFireWallIfNecessary() && pAddress == null) {
      pAddress = findExternalIPwithoutUPnP();
    }
    
    if (pAddress == null) {
      // may need to find and set the firewall
      if (fireWallExternalAddress == null) {
        pAddress = localAddress;
      } else {
         // configure the firewall if necessary, can be any port, start with the freepastry port
        int availableFireWallPort = findAvailableFireWallPort(port, port);  
        openFireWallPort(port, availableFireWallPort);    
        pAddress = new InetSocketAddress(fireWallExternalAddress,availableFireWallPort);                    
      }
    } else {
      // configure the firewall if necessary, but to the specified port
      if (fireWallExternalAddress != null) {
        int availableFireWallPort = findAvailableFireWallPort(port, pAddress.getPort());  
        if (availableFireWallPort == pAddress.getPort()) {
          openFireWallPort(port, availableFireWallPort);
        } else {
          // decide how to handle this           
          switch(getFireWallPolicyVariable("nat_state_policy")) {
            case SocketNATHandler.OVERWRITE:
              openFireWallPort(port, pAddress.getPort());
              break;
            case SocketNATHandler.FAIL:
              // todo: would be useful to pass the app that is bound to that port
              throw new BindException("Firewall is already bound to the requested port:"+pAddress);
            case SocketNATHandler.USE_DIFFERENT_PORT:
              openFireWallPort(port, availableFireWallPort);
              pAddress = new InetSocketAddress(pAddress.getAddress(), availableFireWallPort);
              break;
          }
        }
      }
    }
    return pAddress;
  }

  public boolean test() {
    switch(getFireWallPolicyVariable("nat_test_policy")) {
      case SocketNATHandler.NEVER:
        break;
      case SocketNATHandler.PREFIX_MATCH:
        if (!localAddressIsProbablyNatted()) break;
      case SocketNATHandler.ALWAYS:
      default:  
        ExternalContinuation ec = new ExternalContinuation();
//        pn.testFireWall(bootstrap, ec, 5000);
        ec.sleep();
        Boolean resultB = (Boolean)ec.getResult();
        boolean result = resultB.booleanValue();
        if (result) {
          // continue
        } else { // should change to IOException in FP 2.0
          throw new RuntimeException("Firewall test failed for local:"+localAddress.getAddress()+" external:"+pAddress.getAddress());
        }
    } // switch
    
    return true;
  }

}
